/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit;

import java.io.*;
import java.awt.*;
import java.awt.image.*;

/**
 * Handles loading of bitmap images (.bmp and .pbm) as Images
 * eg. Image im=BitmapLoader.loadImage("pic.bmp");
 * BMP loading code was adapted from the
 * <a href=http://www.wabasoft.com/>Waba</a> bitmap loading source code.
 */
public class BitmapLoader
{
	// Intel architecture getUInt32
	private static int inGetUInt32(byte bytes[], int off)
	{
		return ((bytes[off + 3]&0xFF) << 24) | ((bytes[off + 2]&0xFF) << 16) |
			((bytes[off + 1]&0xFF) << 8) | (bytes[off]&0xFF);
	}

	// Intel architecture getUInt16
	private static int inGetUInt16(byte bytes[], int off)
	{
		return ((bytes[off + 1]&0xFF) << 8) | (bytes[off]&0xFF);
	}

	private static int readBytes(DataInputStream data, byte b[])
	{
		int nread = 0;
		int len = b.length;
		while (true)
		{
			int n = 0;
			try { n = data.read(b, nread, len - nread); }
			catch (Exception e) {e.printStackTrace();}
		  //System.out.println("n="+n);
			if (n <= 0)
				return -1;

			nread += n;
			if (nread == len)
				return len;
		}
	}

	private static void pixelsToRGB(int bitsPerPixel, int width, byte pixels[], int pixelOffset,
		int rgb[], int rgbOffset, int cmap[])
	{
		int mask, step;
		if (bitsPerPixel == 1)
			{
			mask = 0x1;
			step = 1;
			}
		else if (bitsPerPixel == 4)
			{
			mask = 0x0F;
			step = 4;
			}
		else // bitsPerPixel == 8
			{
			mask = 0xFF;
			step = 8;
			}
		int bit = 8 - step;
		int bytnum = pixelOffset;
		int byt = pixels[bytnum++];
		int x = 0;
		while (true)
			{
			int colorIndex = ((mask << bit) & byt) >> bit;
			rgb[rgbOffset++] = cmap[colorIndex];// | (0xFF << 24);
			if (++x >= width)
				break;
			if (bit == 0)
				{
				bit = 8 - step;
				byt = pixels[bytnum++];
				}
			else
				bit -= step;
			}
	}

  private static Image readPBM(DataInputStream data, String name) throws IOException
  {
		boolean binary;
		int width,height;
		int[] rgb=null;
		if (data.read()!='P')
		  return null;
		int type=data.read()-'0';
		if (type==1)
		  binary=false;
		else
		if (type==4)
		  binary=true;
		else
		  return null;

		int outcnt=0;
		int black=Color.black.getRGB();
		int green=Preferences.getIntPref(Bitmap.PREF_BGCOLOR,PilrcEdit.bgColor);
		if (binary)
		{
      int ch;
			while((ch=data.read())<'0'||ch>'9')
			  if (ch=='#')
			    while((ch=data.read())!='\n');

			width=ch-'0';
			while((ch=data.read())>='0'&&ch<='9')
			{
				width*=10;
				width+=ch-'0';
			}
			while((ch=data.read())<'0'||ch>'9');
			height=ch-'0';
			while((ch=data.read())>='0'&&ch<='9')
			{
				height*=10;
				height+=ch-'0';
			}
			rgb=new int[width*height];
			byte[] pixels=new byte[(width+7)/8*height];
			int read=readBytes(data,pixels);
			if (read!=pixels.length)
			{
				System.out.println("read:"+read+" pixels.length="+pixels.length);
			  return null;
			}
			//pixelsToRGB(1,width,pixels,0,rgb,0,new int[]{green,black});
			int incnt=0;
			for(int y=0;y<height;y++)
			  for(int r=0;r<width;r+=8)
			  {
			    byte c=pixels[incnt++];
			    for(int x=0;((x+r)<width)&&(x<8);x++)
			    {
						rgb[outcnt++]=(c&0x80)!=0?black:green;
					  c<<=1;
					}
				}
		}
		else
		{
			StreamTokenizer st=new StreamTokenizer(new InputStreamReader(data));
			st.commentChar('#');
		  st.nextToken();
		  width=(int)st.nval;
		  st.nextToken();
		  height=(int)st.nval;
			rgb=new int[width*height];
			for(int i=0;i<rgb.length;i++)
			{
				st.nextToken();
				rgb[outcnt++]=(int)st.nval==1?black:green;
			}
		}
		// create the image from the RGB buffer
		MemoryImageSource imageSource = new MemoryImageSource(width, height, rgb, 0, width);
		return Toolkit.getDefaultToolkit().createImage(imageSource);
	}

  /**
   * Reads the contents of a .bmp file from the given data stream
   * @param data the stream to read from
   * @param name the name of the filename being read (for error messages)
   * @param type the type of image.  -1 for colour, -2 for gray, or a Bitmap constant (0-3)
   * @param bitmap an optional bitmap where transparency information is read from
   */
	private static Image readBMP(DataInputStream data, String name,int type,Bitmap bitmap)
	{
		// read header (54 bytes)
		// 0-1   magic chars 'BM'
		// 2-5   uint32 filesize (not reliable)
		// 6-7   uint16 0
		// 8-9   uint16 0
		// 10-13 uint32 bitmapOffset
		// 14-17 uint32 info size
		// 18-21 int32  width
		// 22-25 int32  height
		// 26-27 uint16 nplanes
		// 28-29 uint16 bits per pixel
		// 30-33 uint32 compression flag
		// 34-37 uint32 image size in bytes
		// 38-41 int32  biXPelsPerMeter (unused)
		// 32-45 int32  biYPelsPerMeter (unused)
		// 46-49 uint32 colors used (unused)
		// 50-53 uint32 important color count (unused)
		byte header[] = new byte[54];
		if (readBytes(data, header) != 54)
			{
			System.out.println("ERROR: can't read image header for " + name);
			return null;
			}
		if (header[0] != 'B' || header[1] != 'M')
			{
			System.out.println("ERROR: " + name + " is not a BMP image");
			return null;
			}
		int bitmapOffset = inGetUInt32(header, 10);

		int infoSize = inGetUInt32(header, 14);
		if (infoSize != 40)
			{
			System.out.println("ERROR: " + name + " is old-style BMP");
			return null;
			}
		int width = inGetUInt32(header, 18);
		int height = inGetUInt32(header, 22);
		if (width < 0 || height < 0 || width > 65535 || height > 65535)
			{
			System.out.println("ERROR: " + name + " has invalid width/height");
			return null;
			}
		int bpp = inGetUInt16(header, 28);
		if (bpp != 1 && bpp != 4 && bpp != 8)
			{
			System.out.println("ERROR: " + name + " is not a 2, 16 or 256 color image");
			return null;
			}
		int compression = inGetUInt32(header, 30);
		if (compression != 0)
			{
			System.out.println("ERROR: " + name + " is a compressed image");
			return null;
			}
		int numColors = inGetUInt32(header, 46);
		if (numColors==0)
		  numColors=1 << bpp;
		int scanlen = (width * bpp + 7) / 8; // # bytes
		scanlen = ((scanlen + 3) / 4) * 4; // end on 32 bit boundry

		// read colormap
		//
		// 0-3 uint32 col[0]
		// 4-7 uint32 col[1]
		// ...
		int cmapSize = bitmapOffset - 54;
		byte cmapData[] = new byte[cmapSize];

    int bgColor=Preferences.getIntPref(Bitmap.PREF_BGCOLOR);
		if (readBytes(data, cmapData) != cmapSize)
			{
			System.out.println("ERROR: can't read colormap of " + name);
			return null;
			}
		int cmap[] = new int[numColors];
		int j = 0;
		for (int i = 0; i < numColors; i++)
		{
			int blue = cmapData[j++]&0xFF;
			int green = cmapData[j++]&0xFF;
			int red = cmapData[j++]&0xFF;
			j++; // skip reserved

      if (red==green&&green==blue)
      {
        if ((bitmap==null&&type<-1)||
            (bitmap!=null&&(type!=Bitmap.COLOR&&
                           (type!=Bitmap.GREY16||!bitmap.color16))))
        {
          int bg=bgColor;
          blue=blue*(bg&0xFF)/255;
          bg>>=8;
          green=green*(bg&0xFF)/255;
          bg>>=8;
          red=red*(bg&0xFF)/255;
        }
      }
      /*
      else
      if (!color16&&index==Bitmap.GREY16)
      {
        color16=true;
        i=-1; // restart counter
        continue;
      }
      */
      int transparent=255;
      if (type==Bitmap.COLOR&&(i==bitmap.transparentIndex||
           (bitmap.transR==red&&bitmap.transG==green&&bitmap.transB==blue)))
        transparent=0;

			cmap[i] = (transparent << 24) | (red << 16) | (green << 8) | blue;

		}
    /*
    if (index==Bitmap.GREY16)
      bitmap.color16=color16;
      */

		// read pixels and store in RGB buffer
		int rgb[] = new int[width * height];
		byte pixels[] = new byte[scanlen];
		for (int y = height - 1; y >= 0; y--)
			{
			if (readBytes(data, pixels) != scanlen)
				{
				System.out.println("ERROR: scanline " + y + " bad in image " + name);
				return null;
				}
			if (width == 0)
				continue;
			pixelsToRGB(bpp, width, pixels, 0, rgb, y * width, cmap);
			}

		// create the image from the RGB buffer
		MemoryImageSource imageSource = new MemoryImageSource(width, height, rgb, 0, width);
		return Toolkit.getDefaultToolkit().createImage(imageSource);
	}

  /**
   * Loads a bitmap from the jar resources, or a file.
   * @param path the file to load (.bmp or .pbm)
   * @return the loaded image, or null if an error occurs.
   */
	public static Image loadImage(String path)
  {
    return loadImage(path,-1,null);
  }

  /**
   * Loads a bitmap from the jar resources, or a file.
   * @param path the file to load (.bmp or .pbm)
   * @return the loaded image, or null if an error occurs.
   */
  public static Image loadBWImage(String path)
  {
    return loadImage(path,-2,null);
  }

  /**
   * Loads a bitmap from the jar resources, or a file.
   * @param b the bitmap containing the image to load
   * @param type the image type to load from the bitmap
   * @return the loaded image, or null if an error occurs.
   */
  public static Image loadImage(Bitmap b,int type)
  {
    return loadImage(b.getFile(type),type,b);
  }

  /**
   * Loads a bitmap from a file.
   * @param file the file to load the image from
   * @return the loaded image, or null if an error occurs.
   */
  public static Image loadImage(File file)
  {
    return loadImage(file,-1,null);
  }

  private static Image loadImage(String path,int index,Bitmap bitmap)
	{
    try
    {
      InputStream is=BitmapLoader.class.getResourceAsStream(path);
      if (is!=null)
        return loadImage(is,path,index,bitmap);
      else
        return loadImage(new File(path),index,bitmap);
    }
    catch (Exception e)
    {
      System.out.println("Can't load the bitmap: "+path);
    }
    return null;
  }

  private static Image loadImage(File file,int index,Bitmap bitmap)
  {
    try
    {
      InputStream is=new FileInputStream(file);
      return loadImage(is,file.getName(),index,bitmap);
    }
    catch(Exception e)
    {
      System.out.println("Can't load the bitmap: "+file);
    }
    return null;
  }

  /**
   * Loads a bitmap from a file
   * @param file the file the load
   * @return the loaded image, or null if an error occurs.
   */
  private static Image loadImage(InputStream is,String file,int index,Bitmap bitmap)
  {
		try
		{
			DataInputStream dis=new DataInputStream(is);
			String ext=file.substring(file.lastIndexOf('.'));
			if (ext.equalsIgnoreCase(".bmp"))
  			return readBMP(dis,file,index,bitmap);
  	  if (ext.equalsIgnoreCase(".pbm"))
  	    return readPBM(dis,file);
		}
		catch (Exception e)
		{
			System.out.println("Can't load the bitmap: "+file);
      e.printStackTrace();
		};
		return null;
	}
}
