/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.util.Enumeration;
import java.io.*;

/**
 * Handles mouse presses and movements on the screen, sending the appropriate messages
 * to the selected objects and telling the Properties window which object is selected
 */
public class Editor implements ActionListener,MouseListener,MouseMotionListener,KeyListener
{
  private static final int CONTROL_SELECTED=0;
  private static final int FORM_SELECTED=1;
  private static final int KEY_PRESSED=2;

	int zoom;
	Form form;
	FormObject mouseOver;
	int cursor;
	FormObject selected;
	Screen screen;
	Graphics g;
	int sx,sy;
	PopupMenu popup;
	Menu newMenu;
	//Properties props;
	boolean dragging=false;
	boolean editable=true;
	Rectangle dirty=new Rectangle(0,0,-1,0);
  Vector listeners;

  public Editor(Screen screen)
  {
		this.screen=screen;
		screen.addMouseListener(this);
		screen.addMouseMotionListener(this);
		screen.setEditor(this);
		Properties props=new Properties(screen);
    addSelectionListener(props);
		Frame f=new Frame("Properties");
    f.add("North",props);
    f.setBounds(330,0,200,400);
		Preferences.registerFrame(f,true);
		f.setVisible(true);
	}

  public void generatePopupMenu()
  {
		if (popup==null)
		{
		  popup=new PopupMenu();
		  newMenu=new java.awt.Menu("New");
			String[] ot=FormObject.getObjectTypes();
			for(int i=0;i<ot.length;i++)
			{
				MenuItem mi=new MenuItem(ot[i]);
				newMenu.add(mi);
				mi.addActionListener(this);
      }
      screen.add(popup);
		}
		popup.removeAll();
		popup.add(newMenu);
		Vector popups=new Vector();
		if (selected==null)
			form.getPopupMenuItems(popups);
		else
			selected.getPopupMenuItems(popups);
    for(int i=0,size=popups.size();i<size;i++)
    {
			MenuItem mi=new MenuItem((String)popups.elementAt(i));
			mi.addActionListener(this);
      popup.add(mi);
		}
  }

  public void zoomChanged(int zoom)
  {
		this.zoom=zoom;
	}

  public void setForm(Form form)
	{
		this.form=form;
		//props.setFormObject(form);
		selected=null;
	  editable=!(form instanceof Alert);
	  form.setEditor(this);
    notifyListeners(FORM_SELECTED,form);
    notifyListeners(CONTROL_SELECTED,form);
	}

  public Rectangle getDirty()
  {
		return dirty.width==-1?null:dirty;
	}

  public void setDirty(Rectangle r)
  {
		if (r==null)
		  dirty.width=-1;
		else
		if (dirty.width==-1)
		  dirty.setBounds(r.x+form.x,r.y+form.y,r.width,r.height);
		else
		{
			dirty.add(r.x+form.x,r.y+form.y);
			dirty.add(r.x+form.x+r.width,r.y+form.y+r.height);
		}
	}

  /**
   * Process an input string from an external source, such as Drag'n'Drop or
   * the System clipboard.  Input should be considered to be a snippet of .rcp
   * code such as a BUTTON or FORM definition
   * @param s the string to process
   */
  public void processExternalInput(String s)
  {
    processExternalInput(s,-1,-1);
  }

  /**
   * Process an input string from an external source, such as Drag'n'Drop or
   * the System clipboard.  Input should be considered to be a snippet of .rcp
   * code such as a BUTTON or FORM definition
   * @param s the string to process
   * @param x the x coordinate the data arrived from
   * @param y the y coordinate the data arrived from
   */
  public void processExternalInput(String s,int x,int y)
  {
    //System.out.println("process(\""+s+"\","+x+","+y+")");
    BufferedReader br=new BufferedReader(new StringReader(s));
    StreamTokenizer st=new StreamTokenizer(br);
    st.resetSyntax();
    st.whitespaceChars(0,32);
    st.wordChars(33,255);
    st.ordinaryChar('/');
    st.slashSlashComments(true);
    st.lowerCaseMode(true);
    if (form==null||form instanceof Alert)
      return;
    Project p=form.getProject();
    p.getCurrentConstantFile().setAutoGenerate(true);
    FormObject prev=form.getLastObject();
    try
    {
      boolean first=true;
      while(true)
      {
        st.nextToken();
        if (st.ttype==st.TT_EOF)
          break;
        if (st.sval.equals("form"))
        {
          Form nf=new Form(p,st);
          p.addForm(nf,true);
        }
        else
        if (st.sval.equals("alert"))
        {
          Alert na=new Alert(p,st);
          p.addForm(na,true);
        }
        else
        {
          FormObject o=form.parseControl(st,prev,false);
          if (o!=null)
          {
            if (first)
            {
              if (x!=-1&&y!=-1)
                o.setDimensions(""+(x/zoom),""+(y/zoom),null,null);
              setSelected(o);
              first=false;
            }
            prev=o;
          }
          else
            setSelected(null);
        }
      }
      form.clearTriggers();
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
    p.getCurrentConstantFile().setAutoGenerate(false);
  }

  public void paint(Graphics g)
  {
		if (selected!=null&&selected!=form)
		{
			int xa=selected.getXAlignment();
			int ya=selected.getYAlignment();
			int x=form.x+selected.x+(xa==LocationExpression.ALIGN_LEFT?
			  0:selected.width/(xa==LocationExpression.ALIGN_CENTER?2:1));
			int y=form.y+selected.y+(ya==LocationExpression.ALIGN_TOP?
			  0:selected.height/(ya==LocationExpression.ALIGN_CENTER?2:1));
			g.setColor(Color.red);
			g.fillOval(x*zoom-3,y*zoom-3,6,6);
			setDirty(selected.getBounds());
			if (selected.isDependent)
			{
				xa=selected.prev().getXAlignment();
				ya=selected.prev().getYAlignment();
				x=form.x+selected.prev().x+(xa==LocationExpression.ALIGN_LEFT?
				  0:selected.prev().width/(xa==LocationExpression.ALIGN_CENTER?2:1));
				y=form.y+selected.prev().y+(ya==LocationExpression.ALIGN_TOP?
				  0:selected.prev().height/(ya==LocationExpression.ALIGN_CENTER?2:1));
				g.setColor(Color.blue);
				g.fillOval(x*zoom-3,y*zoom-3,6,6);
				setDirty(selected.prev().getBounds());
			}
		}
	}

  public void setSelected(FormObject selected)
  {
		setSelected(selected,true);
	}

  public void setSelected(FormObject selected,boolean repaintAll)
  {
		this.selected=selected;
		//props.setFormObject(selected==null?form:selected);
    notifyListeners(CONTROL_SELECTED,selected==null?form:selected);
		if (repaintAll)
		  screen.repaint();
		else
  		screen.reblat(getDirty());
	}

	/**
	 * Invoked when a key has been typed.
	 * This event occurs when a key press is followed by a key release.
	 */
	public void keyTyped(KeyEvent e) {}

	/**
	 * Invoked when a key has been pressed.
	 */
	public void keyPressed(KeyEvent e)
	{
		switch(e.getKeyCode())
		{
			case KeyEvent.VK_SPACE: screen.toggleBitmap(); break;
      case KeyEvent.VK_TAB:
			  if ((e.getModifiers()&e.VK_SHIFT)>0)
			    setSelected(selected==null||selected.prev()==null?form.lastObject:selected.prev());
			  else
  			  setSelected(selected==null||selected.next()==null?form.firstObject:selected.next());
			  break;
      case KeyEvent.VK_DELETE:
			  if (selected!=null)
			  {
					form.getProject().setDirty(true);
					form.remove(selected);
					setSelected(selected.next());
				}
				break;
		  default:
        notifyListeners(KEY_PRESSED,e);
        /*
		    char ch=e.getKeyChar();
		    if (ch!=e.CHAR_UNDEFINED&&ch>32)
		      props.focusTextField(""+ch);
          */
		    break;
	  }
	}

	/**
	 * Invoked when a key has been released.
	 */
  public void keyReleased(KeyEvent e) {}

	/**
	 * Invoked when an action occurs
	 */
	public void actionPerformed(ActionEvent e)
	{
    form.getProject().setDirty(true);
    String s=e.getActionCommand();

    // check for custom popup menu items.
    if ((selected==null&&form.popupMenuItemSelected(s))||
        (selected!=null&&selected.popupMenuItemSelected(s)))
    {
			screen.repaint();
      return;
		}

    // New <blah>

		if (selected==null)
		{
			FormObject obj=FormObject.createObjectByName(FormObject.convertDisplayNameToObjectType(s));
			if (obj==null)
			  return;
			if (obj.hasLabel())
				obj.label=obj.getObjectType();
			form.add(obj);
			//obj.setDimensions(""+(sx-obj.autoWidth()/2),""+(sy-obj.autoHeight()/2),
			obj.setDimensions(""+sx,""+sy,"AUTO","AUTO");
			form.evaluatePosition();
			setSelected(obj);
		}
		else
		{
			FormObject obj=selected.copy(FormObject.convertDisplayNameToObjectType(s));
			if (obj!=null)
				setSelected(obj);
		}
	}

	/**
	 * Invoked when the mouse has been clicked on a component.
	 */
	public void mouseClicked(MouseEvent e)
	{
		if (e.getClickCount()%2==0&&selected!=null)
		{
		  selected.doubleClicked();
		  screen.repaint();
		}
	}

	/**
	 * Invoked when a mouse button has been pressed on a component.
	 */
	public void mousePressed(MouseEvent e)
	{
		if (editable&&form!=null)
		{
			sx=e.getX()/zoom-form.x;
			sy=e.getY()/zoom-form.y;

			if (mouseOver==null||!mouseOver.isVisible())
			  mouseOver=form.getObjectAt(sx,sy);

		  setSelected(mouseOver,false);
			dragging=true;
		}
		if ((e.getModifiers()&MouseEvent.BUTTON3_MASK)>0)//e.isPopupTrigger())
		{
			generatePopupMenu();
			popup.show(screen,e.getX(),e.getY());
		}
	}

	/**
	 * Invoked when a mouse button has been released on a component.
	 */
	public void mouseReleased(MouseEvent e)
	{
		if (editable&&form!=null)
		{
			dragging=false;
			//screen.repaint();
		}
	}

	/**
	 * Invoked when the mouse enters a component.
	 */
	public void mouseEntered(MouseEvent e) {}

	/**
	 * Invoked when the mouse exits a component.
	 */
	public void mouseExited(MouseEvent e) {}

	/**
	 * Invoked when a mouse button is pressed on a component and then
	 * dragged.  Mouse drag events will continue to be delivered to
	 * the component where the first originated until the mouse button is
	 * released (regardless of whether the mouse position is within the
	 * bounds of the component).
	 */
	public void mouseDragged(MouseEvent e)
	{
  	if (editable&&form!=null&&dragging)
		{
			int ex=e.getX()/zoom-form.x;
			int ey=e.getY()/zoom-form.y;
			int dx=ex-sx;
			int dy=ey-sy;
			if (selected!=null&&(dx!=0||dy!=0))
			{
				if ((selected.isDependent||selected.hasDependent())&&
				  Preferences.getBooleanPref("Auto Disconnect"))
				  selected.disconnect(true);

				form.getProject().setDirty(true);
				int cx=selected.getXAlignment()==LocationExpression.ALIGN_CENTER?dx:0;
				int cy=selected.getYAlignment()==LocationExpression.ALIGN_CENTER?dy:0;
				Rectangle r=null;
				switch(cursor)
				{
					case Cursor.N_RESIZE_CURSOR: selected.transform(0,dy,0,-dy-cy); break;
					case Cursor.NE_RESIZE_CURSOR: selected.transform(-cx,dy,dx+cx,-dy-cy); break;
					case Cursor.E_RESIZE_CURSOR: selected.transform(-cx,0,dx+cx,0); break;
					case Cursor.SE_RESIZE_CURSOR: selected.transform(-cx,-cy,dx+cx,dy+cy); break;
					case Cursor.S_RESIZE_CURSOR: selected.transform(0,-cy,0,dy+cy); break;
					case Cursor.SW_RESIZE_CURSOR: selected.transform(dx,-cy,-dx-cx,dy+cy); break;
					case Cursor.W_RESIZE_CURSOR: selected.transform(dx,0,-dx-cx,0); break;
					case Cursor.NW_RESIZE_CURSOR: selected.transform(dx,dy,-dx-cx,-dy-cy); break;
					case Cursor.MOVE_CURSOR: selected.transform(dx,dy,0,0); break;
				}
			  sx=ex;
			  sy=ey;
			  screen.repaint();
			  PilrcEdit.setStatus(selected.getPositionString());
			}
		}
	}

	/**
	 * Invoked when the mouse button has been moved on a component
	 * (with no buttons no down).
	 */
	public void mouseMoved(MouseEvent e)
	{
		if (editable&&form!=null)
		{
			int x=e.getX()/zoom-form.x;
			int y=e.getY()/zoom-form.y;
			if (selected==null||selected==form||!selected.contains(x,y)||!selected.isVisible())
				mouseOver=form.getObjectAt(x,y);
			else
				mouseOver=selected;
			if (mouseOver!=null&&mouseOver!=form)
			{
				cursor=mouseOver.getCursor(x,y);
				screen.setCursor(Cursor.getPredefinedCursor(mouseOver.getCursor(x,y)));
			}
			else
				screen.setCursor(Cursor.getDefaultCursor());
		}
	}

  public void addSelectionListener(SelectionListener listener)
  {
    if (listeners==null)
      listeners=new Vector();
    listeners.addElement(listener);
  }

  public void removeSelectionListener(SelectionListener listener)
  {
    if (listeners==null)
      listeners=new Vector();
    listeners.removeElement(listener);
  }

  private void notifyListeners(int id,Object data)
  {
    if (listeners==null)
      return;
    for(Enumeration e=listeners.elements();e.hasMoreElements();)
    {
      SelectionListener sl=(SelectionListener)e.nextElement();
      switch(id)
      {
        case CONTROL_SELECTED: sl.controlSelected((FormObject)data); break;
        case FORM_SELECTED: sl.formSelected((Form)data); break;
        case KEY_PRESSED: sl.editorKeyPressed((KeyEvent)data); break;
      }
    }
  }
}