/**
 * PilrcEdit - a graphical editor for PilRC files.
 * Copyright (C) 2000 Rob Nielsen
 * rnielsen@wn.com.au
 * http://www.wn.com.au/rnielsen/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 */
package pilrcedit.control;

import pilrcedit.*;
import java.awt.Graphics;
import java.awt.Color;
import java.io.*;
import java.util.StringTokenizer;

public class Table extends FormObject
{
  public static final int ROWS=0;
  public static final int COLUMNS=1;
  public static final int COLUMNWIDTHS=2;

  public static String PROP_ROWS;
  public static String PROP_COLUMNS;
  public static String PROP_COLUMNWIDTHS;

	Expression rows=new Expression(2);
	Expression cols=new Expression(2);
	Expression[] colWidths;
	String colWidthsString;

  protected Object clone() throws CloneNotSupportedException
	{
		Table t=(Table)super.clone();
		t.rows=new Expression(rows);
		t.cols=new Expression(cols);
		if (colWidths!=null)
		{
			t.colWidths=new Expression[colWidths.length];
			for(int i=0;i<colWidths.length;i++)
				t.colWidths[i]=new Expression(colWidths[i]);
		}

		return t;
	}

  public Table()
  {
    int colSize=autoWidth()/cols.getValue();
    colWidths=new Expression[cols.getValue()];

    for(int i=0;i<colWidths.length;i++)
      colWidths[i]=new Expression(colSize);
  }

  public boolean hasLabel()
  {
		return false;
	}

  public void parse(StreamTokenizer st) throws IOException
	{
    readID(st);
    readLocation(st);
    parseOptions(st);
    evaluatePosition();
	}

  protected boolean parseOption(StreamTokenizer st) throws IOException
  {
		if (super.parseOption(st))
		  return true;

    int key=Keywords.get(this,st.sval);

    switch(key)
    {
      case ROWS: rows.setExpression(readNormalString(st)); break;
      case COLUMNS: cols.setExpression(readNormalString(st)); break;
      case COLUMNWIDTHS:
        int c=cols.getValue();
        colWidths=new Expression[c];
        for(int i=0;i<c;i++)
        {
          colWidths[i]=new Expression(readNormalString(st));
        }
        break;
      default:
        return false;
    }
		return true;
  }

  public String getWidthsAsString()
  {
		if (colWidths==null)
		  return null;
		if (colWidthsString==null)
		{
		  StringBuffer sb=new StringBuffer();
		  sb.append(colWidths[0]);
		  for(int i=1,size=cols.getValue();i<size;i++)
		    sb.append(',').append(colWidths[i]);
		  colWidthsString=sb.toString();
		}
		return colWidthsString;
	}

  public void setColumnWidths(String s)
  {
		colWidthsString=s;
		StringTokenizer st=new StringTokenizer(s,",");
		int newCols=st.countTokens();
		if (colWidths==null||colWidths.length<newCols)
		  colWidths=new Expression[newCols];
		int i=0;
		int sum=0;
		while(st.hasMoreTokens())
		{
			if (colWidths[i]==null)
			  colWidths[i]=new Expression(st.nextToken());
			else
			  colWidths[i].setExpression(st.nextToken());
			sum+=colWidths[i++].getValue();
		}
		cols.setValue(newCols);
		width=sum;
		if (props!=null)
		{
			props.setProperty(PROP_COLUMNS,cols.getExpression());
			props.setProperty(PROP_WIDTH,""+width);
		}
	}

  public int autoWidth()
  {
		if (colWidths==null)
		  return 50;
		int sum=0;
		for(int i=0,size=Math.min(colWidths.length,cols.getValue());i<size;i++)
		  sum+=colWidths[i].getValue();
		return sum;
	}

	public int autoHeight()
	{
		return Font.getFont(0).getHeight()*rows.getValue();
	}

  public void getProperties(Properties prop)
  {
		super.getProperties(prop);
		prop.addProperty(PROP_ROWS,rows.getExpression());
		prop.addProperty(PROP_COLUMNS,cols.getExpression());
		prop.addProperty(PROP_COLUMNWIDTHS,getWidthsAsString(),prop.LIST);
	}

  public boolean setProperty(String name,String value,int index)
  {
		try
		{
			if (name==PROP_ROWS)
			{
				rows.setExpression(value);
				int r=rows.getValue();
				height=height/r*r;
				props.setProperty(PROP_HEIGHT,""+height);
			}
			else
			if (name==PROP_COLUMNS)
      {
				cols.setExpression(value);
      }
			else
			if (name==PROP_COLUMNWIDTHS)
				setColumnWidths(value);
			else
				return super.setProperty(name,value,index);
			evaluatePosition();
			return true;
		}
		catch(NumberFormatException e)
		{
			throw(new IllegalArgumentException());
		}
	}

	protected void writeOptions(StringBuffer sb)
	{
		super.writeOptions(sb);
    Keywords.setDefault(this);
		sb.append(' ').append(Keywords.get(ROWS)).append(' ').append(rows);
		sb.append(' ').append(Keywords.get(COLUMNS)).append(' ').append(cols);
		sb.append(' ').append(Keywords.get(COLUMNWIDTHS));
		for(int i=0,size=cols.getValue();i<size;i++)
		  sb.append(' ').append(colWidths[i]);
	}

  public void paint(Graphics g)
  { 		g.setColor(usable?Color.black:Color.lightGray);
		g.drawRect(x,y,width,height);

		int rowHeight=height/rows.getValue();
		for(int i=rowHeight;i<height;i+=rowHeight)
		  g.drawLine(x,y+i,x+width,y+i);
		if (colWidths!=null)
		{
			int size=Math.min(colWidths.length,cols.getValue())-1;
			for(int i=0,pos=colWidths[0].getValue();i<size;pos+=colWidths[++i].getValue())
				g.drawLine(x+pos,y,x+pos,y+height);
		}
  }
}